@@ -1,4 +1,4 @@
-/* $OpenLDAP: pkg/ldap/libraries/libldap_r/tpool.c,v 1.30.2.17 2007/01/25 13:06:48 hyc Exp $ */
+/* $OpenLDAP: pkg/ldap/libraries/libldap_r/tpool.c,v 1.52.2.7 2007/10/19 01:56:47 hyc Exp $ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
  * Copyright 1998-2007 The OpenLDAP Foundation.
@@ -34,10 +34,10 @@
 typedef enum ldap_int_thread_pool_state_e {
 	LDAP_INT_THREAD_POOL_RUNNING,
 	LDAP_INT_THREAD_POOL_FINISHING,
-	LDAP_INT_THREAD_POOL_STOPPING,
-	LDAP_INT_THREAD_POOL_PAUSING
+	LDAP_INT_THREAD_POOL_STOPPING
 } ldap_int_thread_pool_state_t;
 
+/* Thread-specific key with data and optional free function */
 typedef struct ldap_int_thread_key_s {
 	void *ltk_key;
 	void *ltk_data;
@@ -48,43 +48,76 @@ typedef struct ldap_int_thread_key_s {
  * We don't expect to use many...
  */
 #define	MAXKEYS	32
+
+/* Max number of threads */
 #define	LDAP_MAXTHR	1024	/* must be a power of 2 */
 
-static ldap_pvt_thread_t tid_zero;
+/* Context: thread ID and thread-specific key/data pairs */
+typedef struct ldap_int_thread_userctx_s {
+	ldap_pvt_thread_t ltu_id;
+	ldap_int_thread_key_t ltu_key[MAXKEYS];
+} ldap_int_thread_userctx_t;
 
+
+/* Simple {thread ID -> context} hash table; key=ctx->ltu_id.
+ * Protected by ldap_pvt_thread_pool_mutex except during pauses,
+ * when it is read-only (used by pool_purgekey and pool_context).
+ * Protected by tpool->ltp_mutex during pauses.
+ */
 static struct {
-	ldap_pvt_thread_t id;
-	ldap_int_thread_key_t *ctx;
+	ldap_int_thread_userctx_t *ctx;
+	/* ctx is valid when not NULL or DELETED_THREAD_CTX */
+#	define DELETED_THREAD_CTX (&ldap_int_main_thrctx + 1) /* dummy addr */
 } thread_keys[LDAP_MAXTHR];
-	
 
-typedef struct ldap_int_thread_ctx_s {
+#define	TID_HASH(tid, hash) do { \
+	unsigned const char *ptr_ = (unsigned const char *)&(tid); \
+	unsigned i_; \
+	for (i_ = 0, (hash) = ptr_[0]; ++i_ < sizeof(tid);) \
+		(hash) += ((hash) << 5) ^ ptr_[i_]; \
+} while(0)
+
+
+/* Task for a thread to perform */
+typedef struct ldap_int_thread_task_s {
 	union {
-	LDAP_STAILQ_ENTRY(ldap_int_thread_ctx_s) q;
-	LDAP_SLIST_ENTRY(ldap_int_thread_ctx_s) l;
-	LDAP_SLIST_ENTRY(ldap_int_thread_ctx_s) al;
-	} ltc_next;
-	ldap_pvt_thread_start_t *ltc_start_routine;
-	void *ltc_arg;
-} ldap_int_thread_ctx_t;
+		LDAP_STAILQ_ENTRY(ldap_int_thread_task_s) q;
+		LDAP_SLIST_ENTRY(ldap_int_thread_task_s) l;
+	} ltt_next;
+	ldap_pvt_thread_start_t *ltt_start_routine;
+	void *ltt_arg;
+} ldap_int_thread_task_t;
 
 struct ldap_int_thread_pool_s {
 	LDAP_STAILQ_ENTRY(ldap_int_thread_pool_s) ltp_next;
+
+	/* protect members below, and protect thread_keys[] during pauses */
 	ldap_pvt_thread_mutex_t ltp_mutex;
+
+	/* not paused and something to do for pool_<wrapper/pause/destroy>() */
 	ldap_pvt_thread_cond_t ltp_cond;
+
+	/* ltp_active_count <= 1 && ltp_pause */
 	ldap_pvt_thread_cond_t ltp_pcond;
-	LDAP_STAILQ_HEAD(tcq, ldap_int_thread_ctx_s) ltp_pending_list;
-	LDAP_SLIST_HEAD(tcl, ldap_int_thread_ctx_s) ltp_free_list;
-	LDAP_SLIST_HEAD(tclq, ldap_int_thread_ctx_s) ltp_active_list;
+
+	/* pending tasks, and unused task objects */
+	LDAP_STAILQ_HEAD(tcq, ldap_int_thread_task_s) ltp_pending_list;
+	LDAP_SLIST_HEAD(tcl, ldap_int_thread_task_s) ltp_free_list;
+
 	ldap_int_thread_pool_state_t ltp_state;
-	long ltp_max_count;
-	long ltp_max_pending;
-	long ltp_pending_count;
-	long ltp_active_count;
-	long ltp_open_count;
-	long ltp_starting;
+
+	/* some active request needs to be the sole active request */
+	int ltp_pause;
+
+	long ltp_max_count;			/* max number of threads in pool, or 0 */
+	long ltp_max_pending;		/* max pending or paused requests, or 0 */
+	long ltp_pending_count;		/* pending or paused requests */
+	long ltp_active_count;		/* active, not paused requests */
+	long ltp_open_count;		/* number of threads */
+	long ltp_starting;			/* currenlty starting threads */
 };
 
+static int ldap_int_has_thread_pool = 0;
 static LDAP_STAILQ_HEAD(tpq, ldap_int_thread_pool_s)
 	ldap_int_thread_pool_list =
 	LDAP_STAILQ_HEAD_INITIALIZER(ldap_int_thread_pool_list);
@@ -93,15 +126,13 @@ static ldap_pvt_thread_mutex_t ldap_pvt_thread_pool_mutex;
 
 static void *ldap_int_thread_pool_wrapper( void *pool );
 
-static ldap_pvt_thread_t ldap_int_main_tid;
-
-static ldap_int_thread_key_t ldap_int_main_thrctx[LDAP_MAXTHR];
+/* Context of the main thread */
+static ldap_int_thread_userctx_t ldap_int_main_thrctx;
 
 int
 ldap_int_thread_pool_startup ( void )
 {
-	ldap_int_main_tid = ldap_pvt_thread_self();
-
+	ldap_int_main_thrctx.ltu_id = ldap_pvt_thread_self();
 	return ldap_pvt_thread_mutex_init(&ldap_pvt_thread_pool_mutex);
 }
 
@@ -117,6 +148,8 @@ ldap_int_thread_pool_shutdown ( void )
 	return(0);
 }
 
+
+/* Create a thread pool */
 int
 ldap_pvt_thread_pool_init (
 	ldap_pvt_thread_pool_t *tpool,
@@ -126,6 +159,14 @@ ldap_pvt_thread_pool_init (
 	ldap_pvt_thread_pool_t pool;
 	int rc;
 
+	/* multiple pools are currently not supported (ITS#4943) */
+	assert(!ldap_int_has_thread_pool);
+
+	if (! (0 <= max_threads && max_threads <= LDAP_MAXTHR))
+		max_threads = 0;
+	if (max_pending < 0)
+		max_pending = 0;
+
 	*tpool = NULL;
 	pool = (ldap_pvt_thread_pool_t) LDAP_CALLOC(1,
 		sizeof(struct ldap_int_thread_pool_s));
@@ -141,12 +182,13 @@ ldap_pvt_thread_pool_init (
 	rc = ldap_pvt_thread_cond_init(&pool->ltp_pcond);
 	if (rc != 0)
 		return(rc);
+
+	ldap_int_has_thread_pool = 1;
 	pool->ltp_state = LDAP_INT_THREAD_POOL_RUNNING;
 	pool->ltp_max_count = max_threads;
 	pool->ltp_max_pending = max_pending;
 	LDAP_STAILQ_INIT(&pool->ltp_pending_list);
 	LDAP_SLIST_INIT(&pool->ltp_free_list);
-	LDAP_SLIST_INIT(&pool->ltp_active_list);
 	ldap_pvt_thread_mutex_lock(&ldap_pvt_thread_pool_mutex);
 	LDAP_STAILQ_INSERT_TAIL(&ldap_int_thread_pool_list, pool, ltp_next);
 	ldap_pvt_thread_mutex_unlock(&ldap_pvt_thread_pool_mutex);
@@ -176,6 +218,7 @@ ldap_pvt_thread_pool_init (
 		ldap_pvt_thread_mutex_lock(&ldap_pvt_thread_pool_mutex);
 		LDAP_STAILQ_REMOVE(ldap_int_thread_pool_list, pool, 
 			ldap_int_thread_pool_s, ltp_next);
+		ldap_int_has_thread_pool = 0;
 		ldap_pvt_thread_mutex_unlock(&ldap_pvt_thread_pool_mutex);
 		ldap_pvt_thread_cond_destroy(&pool->ltp_pcond);
 		ldap_pvt_thread_cond_destroy(&pool->ltp_cond);
@@ -189,18 +232,15 @@ ldap_pvt_thread_pool_init (
 	return(0);
 }
 
-#define	TID_HASH(tid, hash) do { unsigned i; \
-	unsigned char *ptr = (unsigned char *)&(tid); \
-	for (i=0, hash=0; i<sizeof(tid); i++) hash += ptr[i]; } while(0)
 
+/* Submit a task to be performed by the thread pool */
 int
 ldap_pvt_thread_pool_submit (
 	ldap_pvt_thread_pool_t *tpool,
 	ldap_pvt_thread_start_t *start_routine, void *arg )
 {
 	struct ldap_int_thread_pool_s *pool;
-	ldap_int_thread_ctx_t *ctx;
-	int need_thread = 0;
+	ldap_int_thread_task_t *task;
 	ldap_pvt_thread_t thr;
 
 	if (tpool == NULL)
@@ -212,108 +252,95 @@ ldap_pvt_thread_pool_submit (
 		return(-1);
 
 	ldap_pvt_thread_mutex_lock(&pool->ltp_mutex);
-	if ((pool->ltp_state != LDAP_INT_THREAD_POOL_RUNNING &&
-		pool->ltp_state != LDAP_INT_THREAD_POOL_PAUSING)
-		|| (pool->ltp_max_pending > 0
+	if (pool->ltp_state != LDAP_INT_THREAD_POOL_RUNNING
+		|| (pool->ltp_max_pending
 			&& pool->ltp_pending_count >= pool->ltp_max_pending))
 	{
 		ldap_pvt_thread_mutex_unlock(&pool->ltp_mutex);
 		return(-1);
 	}
-	ctx = LDAP_SLIST_FIRST(&pool->ltp_free_list);
-	if (ctx) {
-		LDAP_SLIST_REMOVE_HEAD(&pool->ltp_free_list, ltc_next.l);
+
+	task = LDAP_SLIST_FIRST(&pool->ltp_free_list);
+	if (task) {
+		LDAP_SLIST_REMOVE_HEAD(&pool->ltp_free_list, ltt_next.l);
 	} else {
-		ctx = (ldap_int_thread_ctx_t *) LDAP_MALLOC(
-			sizeof(ldap_int_thread_ctx_t));
-		if (ctx == NULL) {
+		task = (ldap_int_thread_task_t *) LDAP_MALLOC(sizeof(*task));
+		if (task == NULL) {
 			ldap_pvt_thread_mutex_unlock(&pool->ltp_mutex);
 			return(-1);
 		}
 	}
 
-	ctx->ltc_start_routine = start_routine;
-	ctx->ltc_arg = arg;
+	task->ltt_start_routine = start_routine;
+	task->ltt_arg = arg;
 
 	pool->ltp_pending_count++;
-	LDAP_STAILQ_INSERT_TAIL(&pool->ltp_pending_list, ctx, ltc_next.q);
-	if (pool->ltp_state == LDAP_INT_THREAD_POOL_PAUSING) {
+	LDAP_STAILQ_INSERT_TAIL(&pool->ltp_pending_list, task, ltt_next.q);
+	if (pool->ltp_pause) {
 		ldap_pvt_thread_mutex_unlock(&pool->ltp_mutex);
 		return(0);
 	}
 	ldap_pvt_thread_cond_signal(&pool->ltp_cond);
 	if (pool->ltp_open_count < pool->ltp_active_count + pool->ltp_pending_count
-		&& (pool->ltp_open_count < pool->ltp_max_count ||
-			pool->ltp_max_count <= 0 ))
+		&& (pool->ltp_open_count <
+			(pool->ltp_max_count ? pool->ltp_max_count : LDAP_MAXTHR)))
 	{
 		pool->ltp_open_count++;
 		pool->ltp_starting++;
-		need_thread = 1;
-	}
-	ldap_pvt_thread_mutex_unlock(&pool->ltp_mutex);
-
-	if (need_thread) {
-		int rc;
-
-		ldap_pvt_thread_mutex_lock(&pool->ltp_mutex);
-
-		rc = ldap_pvt_thread_create( &thr, 1,
-			ldap_int_thread_pool_wrapper, pool );
-		if (rc == 0) {
-			int hash;
-			pool->ltp_starting--;
-
-			/* assign this thread ID to a key slot; start
-			 * at the thread ID itself (mod LDAP_MAXTHR) and
-			 * look for an empty slot.
-			 */
-			TID_HASH(thr, hash);
-			for (rc = hash & (LDAP_MAXTHR-1);
-				!ldap_pvt_thread_equal(thread_keys[rc].id, tid_zero);
-				rc = (rc+1) & (LDAP_MAXTHR-1));
-			thread_keys[rc].id = thr;
-		} else {
+		if (0 != ldap_pvt_thread_create(
+			&thr, 1, ldap_int_thread_pool_wrapper, pool))
+		{
 			/* couldn't create thread.  back out of
 			 * ltp_open_count and check for even worse things.
 			 */
-			pool->ltp_open_count--;
 			pool->ltp_starting--;
+			pool->ltp_open_count--;
 			if (pool->ltp_open_count == 0) {
 				/* no open threads at all?!?
 				 */
-				ldap_int_thread_ctx_t *ptr;
-				LDAP_STAILQ_FOREACH(ptr, &pool->ltp_pending_list, ltc_next.q)
-					if (ptr == ctx) break;
-				if (ptr == ctx) {
-					/* no open threads, context not handled, so
-					 * back out of ltp_pending_count, free the context,
+				ldap_int_thread_task_t *ptr;
+
+				/* let pool_destroy know there are no more threads */
+				ldap_pvt_thread_cond_signal(&pool->ltp_cond);
+
+				LDAP_STAILQ_FOREACH(ptr, &pool->ltp_pending_list, ltt_next.q)
+					if (ptr == task) break;
+				if (ptr == task) {
+					/* no open threads, task not handled, so
+					 * back out of ltp_pending_count, free the task,
 					 * report the error.
 					 */
-					LDAP_STAILQ_REMOVE(&pool->ltp_pending_list, ctx, 
-						ldap_int_thread_ctx_s, ltc_next.q);
-					pool->ltp_pending_count++;
+					LDAP_STAILQ_REMOVE(&pool->ltp_pending_list, task,
+						ldap_int_thread_task_s, ltt_next.q);
+					pool->ltp_pending_count--;
 					ldap_pvt_thread_mutex_unlock(&pool->ltp_mutex);
-					LDAP_FREE(ctx);
+					LDAP_FREE(task);
 					return(-1);
 				}
 			}
 			/* there is another open thread, so this
-			 * context will be handled eventually.
-			 * continue on and signal that the context
-			 * is waiting.
+			 * task will be handled eventually.
+			 * continue on, we have signalled that
+			 * the task is waiting.
 			 */
 		}
-		ldap_pvt_thread_mutex_unlock(&pool->ltp_mutex);
 	}
 
+	ldap_pvt_thread_mutex_unlock(&pool->ltp_mutex);
 	return(0);
 }
 
+/* Set max #threads.  value <= 0 means max supported #threads (LDAP_MAXTHR) */
 int
-ldap_pvt_thread_pool_maxthreads ( ldap_pvt_thread_pool_t *tpool, int max_threads )
+ldap_pvt_thread_pool_maxthreads(
+	ldap_pvt_thread_pool_t *tpool,
+	int max_threads )
 {
 	struct ldap_int_thread_pool_s *pool;
 
+	if (! (0 <= max_threads && max_threads <= LDAP_MAXTHR))
+		max_threads = 0;
+
 	if (tpool == NULL)
 		return(-1);
 
@@ -328,31 +355,133 @@ ldap_pvt_thread_pool_maxthreads ( ldap_pvt_thread_pool_t *tpool, int max_threads
 	return(0);
 }
 
+/* Inspect the pool */
 int
-ldap_pvt_thread_pool_backload ( ldap_pvt_thread_pool_t *tpool )
+ldap_pvt_thread_pool_query(
+	ldap_pvt_thread_pool_t *tpool,
+	ldap_pvt_thread_pool_param_t param,
+	void *value )
 {
-	struct ldap_int_thread_pool_s *pool;
-	int count;
+	struct ldap_int_thread_pool_s	*pool;
+	int				count = -1;
 
-	if (tpool == NULL)
-		return(-1);
+	if ( tpool == NULL || value == NULL ) {
+		return -1;
+	}
 
 	pool = *tpool;
 
-	if (pool == NULL)
-		return(0);
+	if ( pool == NULL ) {
+		return 0;
+	}
 
 	ldap_pvt_thread_mutex_lock(&pool->ltp_mutex);
-	count = pool->ltp_pending_count + pool->ltp_active_count;
-	ldap_pvt_thread_mutex_unlock(&pool->ltp_mutex);
-	return(count);
+	switch ( param ) {
+	case LDAP_PVT_THREAD_POOL_PARAM_MAX:
+		count = pool->ltp_max_count;
+		break;
+
+	case LDAP_PVT_THREAD_POOL_PARAM_MAX_PENDING:
+		count = pool->ltp_max_pending;
+		break;
+
+	case LDAP_PVT_THREAD_POOL_PARAM_OPEN:
+		count = pool->ltp_open_count;
+		break;
+
+	case LDAP_PVT_THREAD_POOL_PARAM_STARTING:
+		count = pool->ltp_starting;
+		break;
+
+	case LDAP_PVT_THREAD_POOL_PARAM_ACTIVE:
+		count = pool->ltp_active_count;
+		break;
+
+	case LDAP_PVT_THREAD_POOL_PARAM_PAUSING:
+		count = pool->ltp_pause;
+		break;
+
+	case LDAP_PVT_THREAD_POOL_PARAM_PENDING:
+		count = pool->ltp_pending_count;
+		break;
+
+	case LDAP_PVT_THREAD_POOL_PARAM_BACKLOAD:
+		count = pool->ltp_pending_count + pool->ltp_active_count;
+		break;
+
+	case LDAP_PVT_THREAD_POOL_PARAM_ACTIVE_MAX:
+		break;
+
+	case LDAP_PVT_THREAD_POOL_PARAM_PENDING_MAX:
+		break;
+
+	case LDAP_PVT_THREAD_POOL_PARAM_BACKLOAD_MAX:
+		break;
+
+	case LDAP_PVT_THREAD_POOL_PARAM_STATE: {
+		static struct {
+			char				*name;
+			ldap_int_thread_pool_state_t	state;
+		}		str2state[] = {
+			{ "running",	LDAP_INT_THREAD_POOL_RUNNING },
+			{ "finishing",	LDAP_INT_THREAD_POOL_FINISHING },
+			{ "stopping",	LDAP_INT_THREAD_POOL_STOPPING },
+			{ NULL }
+		};
+		int		i;
+
+		if ( pool->ltp_pause ) {
+			*((char **)value) = "pausing";
+		} else {
+			for ( i = 0; str2state[ i ].name != NULL; i++ ) {
+				if ( str2state[ i ].state == pool->ltp_state ) {
+					break;
+				}
+			}
+			*((char **)value) = str2state[ i ].name;
+		}
+		if ( *((char **)value) != NULL ) {
+			count = -2;
+		}
+		} break;
+
+	case LDAP_PVT_THREAD_POOL_PARAM_UNKNOWN:
+		break;
+	}
+	ldap_pvt_thread_mutex_unlock( &pool->ltp_mutex );
+
+	if ( count > -1 ) {
+		*((int *)value) = count;
+	}
+
+	return ( count == -1 ? -1 : 0 );
 }
 
+/*
+ * wrapper for ldap_pvt_thread_pool_query(), left around
+ * for backwards compatibility
+ */
+int
+ldap_pvt_thread_pool_backload ( ldap_pvt_thread_pool_t *tpool )
+{
+	int	rc, count;
+
+	rc = ldap_pvt_thread_pool_query( tpool,
+		LDAP_PVT_THREAD_POOL_PARAM_BACKLOAD, (void *)&count );
+
+	if ( rc == 0 ) {
+		return count;
+	}
+
+	return rc;
+}
+
+/* Destroy the pool after making its threads finish */
 int
 ldap_pvt_thread_pool_destroy ( ldap_pvt_thread_pool_t *tpool, int run_pending )
 {
 	struct ldap_int_thread_pool_s *pool, *pptr;
-	ldap_int_thread_ctx_t *ctx;
+	ldap_int_thread_task_t *task;
 
 	if (tpool == NULL)
 		return(-1);
@@ -376,69 +505,86 @@ ldap_pvt_thread_pool_destroy ( ldap_pvt_thread_pool_t *tpool, int run_pending )
 		? LDAP_INT_THREAD_POOL_FINISHING
 		: LDAP_INT_THREAD_POOL_STOPPING;
 
-	if ( pool->ltp_open_count ) {
-		ldap_pvt_thread_cond_broadcast(&pool->ltp_cond);
+	while (pool->ltp_open_count) {
+		if (!pool->ltp_pause)
+			ldap_pvt_thread_cond_broadcast(&pool->ltp_cond);
 		ldap_pvt_thread_cond_wait(&pool->ltp_cond, &pool->ltp_mutex);
 	}
 	ldap_pvt_thread_mutex_unlock(&pool->ltp_mutex);
 
-	while ((ctx = LDAP_STAILQ_FIRST(&pool->ltp_pending_list)) != NULL)
+	while ((task = LDAP_STAILQ_FIRST(&pool->ltp_pending_list)) != NULL)
 	{
-		LDAP_STAILQ_REMOVE_HEAD(&pool->ltp_pending_list, ltc_next.q);
-		LDAP_FREE(ctx);
+		LDAP_STAILQ_REMOVE_HEAD(&pool->ltp_pending_list, ltt_next.q);
+		LDAP_FREE(task);
 	}
 
-	while ((ctx = LDAP_SLIST_FIRST(&pool->ltp_free_list)) != NULL)
+	while ((task = LDAP_SLIST_FIRST(&pool->ltp_free_list)) != NULL)
 	{
-		LDAP_SLIST_REMOVE_HEAD(&pool->ltp_free_list, ltc_next.l);
-		LDAP_FREE(ctx);
+		LDAP_SLIST_REMOVE_HEAD(&pool->ltp_free_list, ltt_next.l);
+		LDAP_FREE(task);
 	}
 
 	ldap_pvt_thread_cond_destroy(&pool->ltp_pcond);
 	ldap_pvt_thread_cond_destroy(&pool->ltp_cond);
 	ldap_pvt_thread_mutex_destroy(&pool->ltp_mutex);
 	LDAP_FREE(pool);
+	*tpool = NULL;
+	ldap_int_has_thread_pool = 0;
 	return(0);
 }
 
+/* Thread loop.  Accept and handle submitted tasks. */
 static void *
 ldap_int_thread_pool_wrapper ( 
 	void *xpool )
 {
 	struct ldap_int_thread_pool_s *pool = xpool;
-	ldap_int_thread_ctx_t *ctx;
-	ldap_int_thread_key_t ltc_key[MAXKEYS];
-	ldap_pvt_thread_t tid;
-	int i, keyslot, hash;
+	ldap_int_thread_task_t *task;
+	ldap_int_thread_userctx_t ctx, *kctx;
+	unsigned i, keyslot, hash;
 
-	if (pool == NULL)
-		return NULL;
+	assert(pool != NULL);
 
 	for ( i=0; i<MAXKEYS; i++ ) {
-		ltc_key[i].ltk_key = NULL;
+		ctx.ltu_key[i].ltk_key = NULL;
 	}
 
-	tid = ldap_pvt_thread_self();
+	ctx.ltu_id = ldap_pvt_thread_self();
+	TID_HASH(ctx.ltu_id, hash);
 
 	ldap_pvt_thread_mutex_lock(&pool->ltp_mutex);
 
-	/* store pointer to our keys */
-	TID_HASH(tid, hash);
-	for (i = hash & (LDAP_MAXTHR-1);
-				!ldap_pvt_thread_equal(thread_keys[i].id, tid);
-				i = (i+1) & (LDAP_MAXTHR-1));
-	thread_keys[i].ctx = ltc_key;
-	keyslot = i;
+	/* thread_keys[] is read-only when paused */
+	while (pool->ltp_pause)
+		ldap_pvt_thread_cond_wait(&pool->ltp_cond, &pool->ltp_mutex);
 
-	while (pool->ltp_state != LDAP_INT_THREAD_POOL_STOPPING) {
-		ctx = LDAP_STAILQ_FIRST(&pool->ltp_pending_list);
-		if (ctx) {
-			LDAP_STAILQ_REMOVE_HEAD(&pool->ltp_pending_list, ltc_next.q);
-		} else {
+	/* find a key slot to give this thread ID and store a
+	 * pointer to our keys there; start at the thread ID
+	 * itself (mod LDAP_MAXTHR) and look for an empty slot.
+	 */
+	ldap_pvt_thread_mutex_lock(&ldap_pvt_thread_pool_mutex);
+	for (keyslot = hash & (LDAP_MAXTHR-1);
+		(kctx = thread_keys[keyslot].ctx) && kctx != DELETED_THREAD_CTX;
+		keyslot = (keyslot+1) & (LDAP_MAXTHR-1));
+	thread_keys[keyslot].ctx = &ctx;
+	ldap_pvt_thread_mutex_unlock(&ldap_pvt_thread_pool_mutex);
+
+	pool->ltp_starting--;
+
+	for (;;) {
+		while (pool->ltp_pause)
+			ldap_pvt_thread_cond_wait(&pool->ltp_cond, &pool->ltp_mutex);
+
+		if (pool->ltp_state == LDAP_INT_THREAD_POOL_STOPPING)
+			break;
+
+		task = LDAP_STAILQ_FIRST(&pool->ltp_pending_list);
+		if (task == NULL) {
 			if (pool->ltp_state == LDAP_INT_THREAD_POOL_FINISHING)
 				break;
-			if (pool->ltp_max_count > 0
-				&& pool->ltp_open_count > pool->ltp_max_count)
+
+			if (pool->ltp_open_count >
+				(pool->ltp_max_count ? pool->ltp_max_count : LDAP_MAXTHR))
 			{
 				/* too many threads running (can happen if the
 				 * maximum threads value is set during ongoing
@@ -454,50 +600,41 @@ ldap_int_thread_pool_wrapper (
 			 * always have at least one thread open).  the check
 			 * should be like this:
 			 *   if (pool->ltp_open_count > 1 && pool->ltp_starting == 0)
-			 *       check timer, leave thread (break;)
+			 *       check timer, wait if ltp_pause, leave thread (break;)
 			 *
 			 * Just use pthread_cond_timedwait if we want to
 			 * check idle time.
 			 */
 
-			if (pool->ltp_state == LDAP_INT_THREAD_POOL_RUNNING
-				|| pool->ltp_state == LDAP_INT_THREAD_POOL_PAUSING)
-			{
-				ldap_pvt_thread_cond_wait(&pool->ltp_cond, &pool->ltp_mutex);
-			}
-
+			assert(pool->ltp_state == LDAP_INT_THREAD_POOL_RUNNING);
+			ldap_pvt_thread_cond_wait(&pool->ltp_cond, &pool->ltp_mutex);
 			continue;
 		}
 
+		LDAP_STAILQ_REMOVE_HEAD(&pool->ltp_pending_list, ltt_next.q);
 		pool->ltp_pending_count--;
-
-		LDAP_SLIST_INSERT_HEAD(&pool->ltp_active_list, ctx, ltc_next.al);
 		pool->ltp_active_count++;
 		ldap_pvt_thread_mutex_unlock(&pool->ltp_mutex);
 
-		ctx->ltc_start_routine(ltc_key, ctx->ltc_arg);
+		task->ltt_start_routine(&ctx, task->ltt_arg);
 
 		ldap_pvt_thread_mutex_lock(&pool->ltp_mutex);
-		LDAP_SLIST_REMOVE(&pool->ltp_active_list, ctx,
-			ldap_int_thread_ctx_s, ltc_next.al);
-		LDAP_SLIST_INSERT_HEAD(&pool->ltp_free_list, ctx, ltc_next.l);
+		LDAP_SLIST_INSERT_HEAD(&pool->ltp_free_list, task, ltt_next.l);
 		pool->ltp_active_count--;
-
-		if (pool->ltp_state == LDAP_INT_THREAD_POOL_PAUSING) {
-			if (pool->ltp_active_count < 2) {
-				ldap_pvt_thread_cond_signal(&pool->ltp_pcond);
-			}
-			ldap_pvt_thread_cond_wait(&pool->ltp_cond, &pool->ltp_mutex);
-		}
+		/* let pool_pause know when it is the sole active thread */
+		if (pool->ltp_active_count < 2)
+			ldap_pvt_thread_cond_signal(&pool->ltp_pcond);
 	}
 
-	ldap_pvt_thread_pool_context_reset( ltc_key );
+	/* The ltp_mutex lock protects ctx->ltu_key from pool_purgekey()
+	 * during this call, since it prevents new pauses. */
+	ldap_pvt_thread_pool_context_reset(&ctx);
 
-	thread_keys[keyslot].ctx = NULL;
-	thread_keys[keyslot].id = tid_zero;
+	ldap_pvt_thread_mutex_lock(&ldap_pvt_thread_pool_mutex);
+	thread_keys[keyslot].ctx = DELETED_THREAD_CTX;
+	ldap_pvt_thread_mutex_unlock(&ldap_pvt_thread_pool_mutex);
 
 	pool->ltp_open_count--;
-
 	/* let pool_destroy know we're all done */
 	if (pool->ltp_open_count < 1)
 		ldap_pvt_thread_cond_signal(&pool->ltp_cond);
@@ -508,6 +645,7 @@ ldap_int_thread_pool_wrapper (
 	return(NULL);
 }
 
+/* Pause the pool.  Return when all other threads are paused. */
 int
 ldap_pvt_thread_pool_pause ( 
 	ldap_pvt_thread_pool_t *tpool )
@@ -525,22 +663,30 @@ ldap_pvt_thread_pool_pause (
 	ldap_pvt_thread_mutex_lock(&pool->ltp_mutex);
 
 	/* If someone else has already requested a pause, we have to wait */
-	while (pool->ltp_state == LDAP_INT_THREAD_POOL_PAUSING) {
+	if (pool->ltp_pause) {
 		pool->ltp_pending_count++;
 		pool->ltp_active_count--;
-		ldap_pvt_thread_cond_wait(&pool->ltp_cond, &pool->ltp_mutex);
+		/* let the other pool_pause() know when it can proceed */
+		if (pool->ltp_active_count < 2)
+			ldap_pvt_thread_cond_signal(&pool->ltp_pcond);
+		do {
+			ldap_pvt_thread_cond_wait(&pool->ltp_cond, &pool->ltp_mutex);
+		} while (pool->ltp_pause);
 		pool->ltp_pending_count--;
 		pool->ltp_active_count++;
 	}
-	/* Wait for everyone else to finish */
-	pool->ltp_state = LDAP_INT_THREAD_POOL_PAUSING;
+
+	/* Wait for everyone else to pause or finish */
+	pool->ltp_pause = 1;
 	while (pool->ltp_active_count > 1) {
 		ldap_pvt_thread_cond_wait(&pool->ltp_pcond, &pool->ltp_mutex);
 	}
+
 	ldap_pvt_thread_mutex_unlock(&pool->ltp_mutex);
 	return(0);
 }
 
+/* End a pause */
 int
 ldap_pvt_thread_pool_resume ( 
 	ldap_pvt_thread_pool_t *tpool )
@@ -556,57 +702,85 @@ ldap_pvt_thread_pool_resume (
 		return(0);
 
 	ldap_pvt_thread_mutex_lock(&pool->ltp_mutex);
-
-	pool->ltp_state = LDAP_INT_THREAD_POOL_RUNNING;
-	ldap_pvt_thread_cond_broadcast(&pool->ltp_cond);
+	pool->ltp_pause = 0;
+	if (pool->ltp_state == LDAP_INT_THREAD_POOL_RUNNING)
+		ldap_pvt_thread_cond_broadcast(&pool->ltp_cond);
 	ldap_pvt_thread_mutex_unlock(&pool->ltp_mutex);
 	return(0);
 }
 
+/*
+ * Get the key's data and optionally free function in the given context.
+ */
 int ldap_pvt_thread_pool_getkey(
 	void *xctx,
 	void *key,
 	void **data,
 	ldap_pvt_thread_pool_keyfree_t **kfree )
 {
-	ldap_int_thread_key_t *ctx = xctx;
+	ldap_int_thread_userctx_t *ctx = xctx;
 	int i;
 
-	if ( !ctx || !data ) return EINVAL;
+	if ( !ctx || !key || !data ) return EINVAL;
 
-	for ( i=0; i<MAXKEYS && ctx[i].ltk_key; i++ ) {
-		if ( ctx[i].ltk_key == key ) {
-			*data = ctx[i].ltk_data;
-			if ( kfree ) *kfree = ctx[i].ltk_free;
+	for ( i=0; i<MAXKEYS && ctx->ltu_key[i].ltk_key; i++ ) {
+		if ( ctx->ltu_key[i].ltk_key == key ) {
+			*data = ctx->ltu_key[i].ltk_data;
+			if ( kfree ) *kfree = ctx->ltu_key[i].ltk_free;
 			return 0;
 		}
 	}
 	return ENOENT;
 }
 
+static void
+clear_key_idx( ldap_int_thread_userctx_t *ctx, int i )
+{
+	for ( ; i < MAXKEYS-1 && ctx->ltu_key[i+1].ltk_key; i++ )
+		ctx->ltu_key[i] = ctx->ltu_key[i+1];
+	ctx->ltu_key[i].ltk_key = NULL;
+}
+
+/*
+ * Set or remove data for the key in the given context.
+ * key can be any unique pointer.
+ * kfree() is an optional function to free the data (but not the key):
+ *   pool_context_reset() and pool_purgekey() call kfree(key, data),
+ *   but pool_setkey() does not.  For pool_setkey() it is the caller's
+ *   responsibility to free any existing data with the same key.
+ *   kfree() must not call functions taking a tpool argument.
+ */
 int ldap_pvt_thread_pool_setkey(
 	void *xctx,
 	void *key,
 	void *data,
 	ldap_pvt_thread_pool_keyfree_t *kfree )
 {
-	ldap_int_thread_key_t *ctx = xctx;
-	int i;
+	ldap_int_thread_userctx_t *ctx = xctx;
+	int i, found;
 
 	if ( !ctx || !key ) return EINVAL;
 
-	for ( i=0; i<MAXKEYS; i++ ) {
-		if ( !ctx[i].ltk_key || ctx[i].ltk_key == key ) {
-			if ( data || kfree )
-				ctx[i].ltk_key = key;
-			else
-				ctx[i].ltk_key = NULL;
-			ctx[i].ltk_data = data;
-			ctx[i].ltk_free = kfree;
-			return 0;
+	for ( i=found=0; i<MAXKEYS; i++ ) {
+		if ( ctx->ltu_key[i].ltk_key == key ) {
+			found = 1;
+			break;
+		} else if ( !ctx->ltu_key[i].ltk_key ) {
+			break;
 		}
 	}
-	return ENOMEM;
+
+	if ( data || kfree ) {
+		if ( i>=MAXKEYS )
+			return ENOMEM;
+		ctx->ltu_key[i].ltk_key = key;
+		ctx->ltu_key[i].ltk_data = data;
+		ctx->ltu_key[i].ltk_free = kfree;
+	} else if ( found ) {
+		clear_key_idx( ctx, i );
+	}
+
+	return 0;
 }
 
 /* Free all elements with this key, no matter which thread they're in.
@@ -615,17 +789,19 @@ int ldap_pvt_thread_pool_setkey(
 void ldap_pvt_thread_pool_purgekey( void *key )
 {
 	int i, j;
-	ldap_int_thread_key_t *ctx;
+	ldap_int_thread_userctx_t *ctx;
+
+	assert ( key != NULL );
 
 	for ( i=0; i<LDAP_MAXTHR; i++ ) {
-		if ( thread_keys[i].ctx ) {
-			ctx = thread_keys[i].ctx;
-			for ( j=0; j<MAXKEYS; j++ ) {
-				if ( ctx[j].ltk_key == key ) {
-					if (ctx[j].ltk_free)
-						ctx[j].ltk_free( ctx[j].ltk_key, ctx[j].ltk_data );
-					ctx[j].ltk_key = NULL;
-					ctx[j].ltk_free = NULL;
+		ctx = thread_keys[i].ctx;
+		if ( ctx && ctx != DELETED_THREAD_CTX ) {
+			for ( j=0; j<MAXKEYS && ctx->ltu_key[j].ltk_key; j++ ) {
+				if ( ctx->ltu_key[j].ltk_key == key ) {
+					if (ctx->ltu_key[j].ltk_free)
+						ctx->ltu_key[j].ltk_free( ctx->ltu_key[j].ltk_key,
+						ctx->ltu_key[j].ltk_data );
+					clear_key_idx( ctx, j );
 					break;
 				}
 			}
@@ -634,41 +810,64 @@ void ldap_pvt_thread_pool_purgekey( void *key )
 }
 
 /*
+ * Find the context of the current thread.
  * This is necessary if the caller does not have access to the
  * thread context handle (for example, a slapd plugin calling
- * slapi_search_internal()). No doubt it is more efficient to
+ * slapi_search_internal()). No doubt it is more efficient
  * for the application to keep track of the thread context
  * handles itself.
  */
 void *ldap_pvt_thread_pool_context( )
 {
 	ldap_pvt_thread_t tid;
-	int i, hash;
+	unsigned i, hash;
+	ldap_int_thread_userctx_t *ctx;
 
 	tid = ldap_pvt_thread_self();
-	if ( ldap_pvt_thread_equal( tid, ldap_int_main_tid ))
-		return ldap_int_main_thrctx;
+	if ( ldap_pvt_thread_equal( tid, ldap_int_main_thrctx.ltu_id ))
+		return &ldap_int_main_thrctx;
 
 	TID_HASH( tid, hash );
-	for (i = hash & (LDAP_MAXTHR-1);
-		!ldap_pvt_thread_equal(thread_keys[i].id, tid_zero) &&
-		!ldap_pvt_thread_equal(thread_keys[i].id, tid);
-		i = (i+1) & (LDAP_MAXTHR-1));
+	i = hash &= (LDAP_MAXTHR-1);
+	ldap_pvt_thread_mutex_lock(&ldap_pvt_thread_pool_mutex);
+	do {
+		ctx = thread_keys[i].ctx;
+		if ( ctx == DELETED_THREAD_CTX )
+			continue;
+		if ( !ctx || ldap_pvt_thread_equal(thread_keys[i].ctx->ltu_id, tid) )
+			goto done;
+	} while ( (i = (i+1) & (LDAP_MAXTHR-1)) != hash );
+	ctx = NULL;
+ done:
+	ldap_pvt_thread_mutex_unlock(&ldap_pvt_thread_pool_mutex);
 
-	return thread_keys[i].ctx;
+	return ctx;
 }
 
+/*
+ * Free the context's keys.
+ * Must not call functions taking a tpool argument (because this
+ * thread already holds ltp_mutex when called from pool_wrapper()).
+ */
 void ldap_pvt_thread_pool_context_reset( void *vctx )
 {
-	ldap_int_thread_key_t *ctx = vctx;
+	ldap_int_thread_userctx_t *ctx = vctx;
 	int i;
 
 	for ( i=MAXKEYS-1; i>=0; i--) {
-		if ( ctx[i].ltk_key )
+		if ( !ctx->ltu_key[i].ltk_key )
 			continue;
-		if ( ctx[i].ltk_free )
-			ctx[i].ltk_free( ctx[i].ltk_key, ctx[i].ltk_data );
-		ctx[i].ltk_key = NULL;
+		if ( ctx->ltu_key[i].ltk_free )
+			ctx->ltu_key[i].ltk_free( ctx->ltu_key[i].ltk_key,
+			ctx->ltu_key[i].ltk_data );
+		ctx->ltu_key[i].ltk_key = NULL;
 	}
 }
+
+ldap_pvt_thread_t ldap_pvt_thread_pool_tid( void *vctx )
+{
+	ldap_int_thread_userctx_t *ctx = vctx;
+
+	return ctx->ltu_id;
+}
 #endif /* LDAP_THREAD_HAVE_TPOOL */
