@@ -1,8 +1,8 @@
 /* dn2id.c - routines to deal with the dn2id index */
-/* $OpenLDAP: pkg/ldap/servers/slapd/back-bdb/dn2id.c,v 1.106.2.18 2008/02/11 23:24:19 kurt Exp $ */
+/* $OpenLDAP: pkg/ldap/servers/slapd/back-bdb/dn2id.c,v 1.137.2.2 2007/08/31 23:14:01 quanah Exp $ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 2000-2008 The OpenLDAP Foundation.
+ * Copyright 2000-2007 The OpenLDAP Foundation.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -23,32 +23,6 @@
 #include "idl.h"
 #include "lutil.h"
 
-#define bdb_dn2id_lock					BDB_SYMBOL(dn2id_lock)
-
-static int
-bdb_dn2id_lock( struct bdb_info *bdb, struct berval *dn,
-	int rw, u_int32_t locker, DB_LOCK *lock )
-{
-	int       rc;
-	DBT       lockobj;
-	int       db_rw;
-
-	if (!locker)
-		return 0;
-
-	if (rw)
-		db_rw = DB_LOCK_WRITE;
-	else
-		db_rw = DB_LOCK_READ;
-
-	lockobj.data = dn->bv_val;
-	lockobj.size = dn->bv_len;
-
-	rc = LOCK_GET(bdb->bi_dbenv, locker, DB_LOCK_NOWAIT,
-					&lockobj, db_rw, lock);
-	return rc;
-}
-
 #ifndef BDB_HIER
 int
 bdb_dn2id_add(
@@ -65,8 +39,8 @@ bdb_dn2id_add(
 	char		*buf;
 	struct berval	ptr, pdn;
 
-	Debug( LDAP_DEBUG_TRACE, "=> bdb_dn2id_add 0x%lx: \"%s\"\n",
-		e->e_id, e->e_ndn, 0 );
+	Debug( LDAP_DEBUG_TRACE, "=> bdb_dn2id_add( \"%s\", 0x%08lx )\n",
+		e->e_ndn, (long) e->e_id, 0 );
 	assert( e->e_id != NOID );
 
 	DBTzero( &key );
@@ -89,8 +63,8 @@ bdb_dn2id_add(
 	/* store it -- don't override */
 	rc = db->put( db, txn, &key, &data, DB_NOOVERWRITE );
 	if( rc != 0 ) {
-		Debug( LDAP_DEBUG_ANY, "=> bdb_dn2id_add 0x%lx: put failed: %s %d\n",
-			e->e_id, db_strerror(rc), rc );
+		Debug( LDAP_DEBUG_ANY, "=> bdb_dn2id_add: put failed: %s %d\n",
+			db_strerror(rc), rc, 0 );
 		goto done;
 	}
 
@@ -102,8 +76,8 @@ bdb_dn2id_add(
 		rc = db->put( db, txn, &key, &data, DB_NOOVERWRITE );
 		if( rc != 0 ) {
 			Debug( LDAP_DEBUG_ANY,
-			"=> bdb_dn2id_add 0x%lx: subtree (%s) put failed: %d\n",
-			e->e_id, ptr.bv_val, rc );
+			"=> bdb_dn2id_add: subtree (%s) put failed: %d\n",
+			ptr.bv_val, rc, 0 );
 			goto done;
 		}
 		
@@ -123,8 +97,8 @@ bdb_dn2id_add(
 
 		if( rc != 0 ) {
 			Debug( LDAP_DEBUG_ANY,
-				"=> bdb_dn2id_add 0x%lx: parent (%s) insert failed: %d\n",
-					e->e_id, ptr.bv_val, rc );
+				"=> bdb_dn2id_add: parent (%s) insert failed: %d\n",
+					ptr.bv_val, rc, 0 );
 			goto done;
 		}
 	}
@@ -141,8 +115,8 @@ bdb_dn2id_add(
 
 		if( rc != 0 ) {
 			Debug( LDAP_DEBUG_ANY,
-				"=> bdb_dn2id_add 0x%lx: subtree (%s) insert failed: %d\n",
-					e->e_id, ptr.bv_val, rc );
+				"=> bdb_dn2id_add: subtree (%s) insert failed: %d\n",
+					ptr.bv_val, rc, 0 );
 			break;
 		}
 #ifdef BDB_MULTIPLE_SUFFIXES
@@ -159,7 +133,7 @@ bdb_dn2id_add(
 
 done:
 	op->o_tmpfree( buf, op->o_tmpmemctx );
-	Debug( LDAP_DEBUG_TRACE, "<= bdb_dn2id_add 0x%lx: %d\n", e->e_id, rc, 0 );
+	Debug( LDAP_DEBUG_TRACE, "<= bdb_dn2id_add: %d\n", rc, 0, 0 );
 	return rc;
 }
 
@@ -172,14 +146,13 @@ bdb_dn2id_delete(
 {
 	struct bdb_info *bdb = (struct bdb_info *) op->o_bd->be_private;
 	DB *db = bdb->bi_dn2id->bdi_db;
-	char		*buf;
-	DBT		key;
-	DB_LOCK	lock;
-	struct berval	pdn, ptr;
 	int		rc;
+	DBT		key;
+	char		*buf;
+	struct berval	pdn, ptr;
 
-	Debug( LDAP_DEBUG_TRACE, "=> bdb_dn2id_delete 0x%lx: \"%s\"\n",
-		e->e_id, e->e_ndn, 0 );
+	Debug( LDAP_DEBUG_TRACE, "=> bdb_dn2id_delete( \"%s\", 0x%08lx )\n",
+		e->e_ndn, e->e_id, 0 );
 
 	DBTzero( &key );
 	key.size = e->e_nname.bv_len + 2;
@@ -192,15 +165,11 @@ bdb_dn2id_delete(
 	AC_MEMCPY( ptr.bv_val, e->e_nname.bv_val, e->e_nname.bv_len );
 	ptr.bv_val[ptr.bv_len] = '\0';
 
-	/* We hold this lock until the TXN completes */
-	rc = bdb_dn2id_lock( bdb, &e->e_nname, 1, TXN_ID( txn ), &lock );
-	if ( rc ) goto done;
-
 	/* delete it */
 	rc = db->del( db, txn, &key, 0 );
 	if( rc != 0 ) {
-		Debug( LDAP_DEBUG_ANY, "=> bdb_dn2id_delete 0x%lx: delete failed: %s %d\n",
-			e->e_id, db_strerror(rc), rc );
+		Debug( LDAP_DEBUG_ANY, "=> bdb_dn2id_delete: delete failed: %s %d\n",
+			db_strerror(rc), rc, 0 );
 		goto done;
 	}
 
@@ -212,8 +181,8 @@ bdb_dn2id_delete(
 		rc = bdb_idl_delete_key( op->o_bd, db, txn, &key, e->e_id );
 		if( rc != 0 ) {
 			Debug( LDAP_DEBUG_ANY,
-			"=> bdb_dn2id_delete 0x%lx: subtree (%s) delete failed: %d\n",
-			e->e_id, ptr.bv_val, rc );
+			"=> bdb_dn2id_delete: subtree (%s) delete failed: %d\n",
+			ptr.bv_val, rc, 0 );
 			goto done;
 		}
 
@@ -233,8 +202,8 @@ bdb_dn2id_delete(
 
 		if( rc != 0 ) {
 			Debug( LDAP_DEBUG_ANY,
-				"=> bdb_dn2id_delete 0x%lx: parent (%s) delete failed: %d\n",
-				e->e_id, ptr.bv_val, rc );
+				"=> bdb_dn2id_delete: parent (%s) delete failed: %d\n",
+				ptr.bv_val, rc, 0 );
 			goto done;
 		}
 	}
@@ -250,8 +219,8 @@ bdb_dn2id_delete(
 		rc = bdb_idl_delete_key( op->o_bd, db, txn, &key, e->e_id );
 		if( rc != 0 ) {
 			Debug( LDAP_DEBUG_ANY,
-				"=> bdb_dn2id_delete 0x%lx: subtree (%s) delete failed: %d\n",
-				e->e_id, ptr.bv_val, rc );
+				"=> bdb_dn2id_delete: subtree (%s) delete failed: %d\n",
+				ptr.bv_val, rc, 0 );
 			goto done;
 		}
 #ifdef BDB_MULTIPLE_SUFFIXES
@@ -268,27 +237,24 @@ bdb_dn2id_delete(
 
 done:
 	op->o_tmpfree( buf, op->o_tmpmemctx );
-	Debug( LDAP_DEBUG_TRACE, "<= bdb_dn2id_delete 0x%lx: %d\n", e->e_id, rc, 0 );
+	Debug( LDAP_DEBUG_TRACE, "<= bdb_dn2id_delete %d\n", rc, 0, 0 );
 	return rc;
 }
 
 int
 bdb_dn2id(
 	Operation *op,
+	DB_TXN *txn,
 	struct berval	*dn,
-	EntryInfo *ei,
-	u_int32_t locker,
-	DB_LOCK *lock )
+	EntryInfo *ei )
 {
 	struct bdb_info *bdb = (struct bdb_info *) op->o_bd->be_private;
 	DB *db = bdb->bi_dn2id->bdi_db;
-	DBC	*cursor;
 	int		rc;
 	DBT		key, data;
 	ID		nid;
 
 	Debug( LDAP_DEBUG_TRACE, "=> bdb_dn2id(\"%s\")\n", dn->bv_val, 0, 0 );
-
 	DBTzero( &key );
 	key.size = dn->bv_len + 2;
 	key.data = op->o_tmpalloc( key.size, op->o_tmpmemctx );
@@ -301,31 +267,18 @@ bdb_dn2id(
 	data.ulen = sizeof(ID);
 	data.flags = DB_DBT_USERMEM;
 
-	rc = db->cursor( db, NULL, &cursor, bdb->bi_db_opflags );
-	if ( rc ) goto leave;
-
-	rc = bdb_dn2id_lock( bdb, dn, 0, locker, lock );
-	if ( rc ) goto nolock;
-
-	if ( locker ) {
-		cursor->locker = locker;
-	}
-
 	/* fetch it */
-	rc = cursor->c_get( cursor, &key, &data, DB_SET );
-
-nolock:
-	cursor->c_close( cursor );
-leave:
+	rc = db->get( db, txn, &key, &data, bdb->bi_db_opflags );
 
 	if( rc != 0 ) {
 		Debug( LDAP_DEBUG_TRACE, "<= bdb_dn2id: get failed: %s (%d)\n",
 			db_strerror( rc ), rc, 0 );
 	} else {
 		BDB_DISK2ID( &nid, &ei->bei_id );
-		Debug( LDAP_DEBUG_TRACE, "<= bdb_dn2id: got id=0x%lx\n",
+		Debug( LDAP_DEBUG_TRACE, "<= bdb_dn2id: got id=0x%08lx\n",
 			ei->bei_id, 0, 0 );
 	}
+
 	op->o_tmpfree( key.data, op->o_tmpmemctx );
 	return rc;
 }
@@ -379,7 +332,9 @@ bdb_dn2id_children(
 int
 bdb_dn2idl(
 	Operation *op,
-	Entry *e,
+	BDB_LOCKER locker,
+	struct berval *ndn,
+	EntryInfo *ei,
 	ID *ids,
 	ID *stack )
 {
@@ -391,25 +346,26 @@ bdb_dn2idl(
 		? DN_ONE_PREFIX : DN_SUBTREE_PREFIX;
 
 	Debug( LDAP_DEBUG_TRACE, "=> bdb_dn2idl(\"%s\")\n",
-		e->e_nname.bv_val, 0, 0 );
+		ndn->bv_val, 0, 0 );
 
 #ifndef	BDB_MULTIPLE_SUFFIXES
-	if ( prefix == DN_SUBTREE_PREFIX && BEI(e)->bei_parent->bei_id == 0 ) {
+	if ( prefix == DN_SUBTREE_PREFIX
+		&& ( ei->bei_id == 0 || ei->bei_parent->bei_id == 0 )) {
 		BDB_IDL_ALL(bdb, ids);
 		return 0;
 	}
 #endif
 
 	DBTzero( &key );
-	key.size = e->e_nname.bv_len + 2;
+	key.size = ndn->bv_len + 2;
 	key.ulen = key.size;
 	key.flags = DB_DBT_USERMEM;
 	key.data = op->o_tmpalloc( key.size, op->o_tmpmemctx );
 	((char *)key.data)[0] = prefix;
-	AC_MEMCPY( &((char *)key.data)[1], e->e_nname.bv_val, key.size - 1 );
+	AC_MEMCPY( &((char *)key.data)[1], ndn->bv_val, key.size - 1 );
 
 	BDB_IDL_ZERO( ids );
-	rc = bdb_idl_fetch_key( op->o_bd, db, NULL, &key, ids, NULL, 0 );
+	rc = bdb_idl_fetch_key( op->o_bd, db, locker, &key, ids, NULL, 0 );
 
 	if( rc != 0 ) {
 		Debug( LDAP_DEBUG_TRACE,
@@ -451,31 +407,6 @@ typedef struct diskNode {
 	unsigned char entryID[sizeof(ID)];  /* variable placement */
 } diskNode;
 
-/* Sort function for the sorted duplicate data items of a dn2id key.
- * Sorts based on normalized RDN, in length order.
- */
-int
-hdb_dup_compare(
-	DB *db, 
-	const DBT *usrkey,
-	const DBT *curkey
-)
-{
-	diskNode *un, *cn;
-	int rc, ul, cl;
-
-	un = (diskNode *)usrkey->data;
-	cn = (diskNode *)curkey->data;
-
-	/* data is not aligned, cannot compare directly */
-	rc = un->nrdnlen[0] - cn->nrdnlen[0];
-	if ( rc ) return rc;
-	rc = un->nrdnlen[1] - cn->nrdnlen[1];
-	if ( rc ) return rc;
-
-	return strcmp( un->nrdn, cn->nrdn );
-}
-
 /* This function constructs a full DN for a given entry.
  */
 int hdb_fix_dn(
@@ -551,9 +482,6 @@ hdb_dn2id_add(
 	diskNode *d;
 	char *ptr;
 
-	Debug( LDAP_DEBUG_TRACE, "=> hdb_dn2id_add 0x%lx: \"%s\"\n",
-		e->e_id, e->e_ndn, 0 );
-
 	nrlen = dn_rdnlen( op->o_bd, &e->e_nname );
 	if (nrlen) {
 		rlen = dn_rdnlen( op->o_bd, &e->e_name );
@@ -620,10 +548,7 @@ hdb_dn2id_add(
 			bdb_idl_cache_add_id( bdb, db, &key, e->e_id );
 		}
 	}
-
-leave:
 	op->o_tmpfree( d, op->o_tmpmemctx );
-	Debug( LDAP_DEBUG_TRACE, "<= hdb_dn2id_add 0x%lx: %d\n", e->e_id, rc, 0 );
 
 	return rc;
 }
@@ -643,10 +568,6 @@ hdb_dn2id_delete(
 	int rc;
 	ID	nid;
 	unsigned char dlen[2];
-	DB_LOCK	lock;
-
-	Debug( LDAP_DEBUG_TRACE, "=> hdb_dn2id_delete 0x%lx: \"%s\"\n",
-		e->e_id, e->e_ndn, 0 );
 
 	DBTzero(&key);
 	key.size = sizeof(ID);
@@ -661,6 +582,8 @@ hdb_dn2id_delete(
 	data.flags = DB_DBT_USERMEM | DB_DBT_PARTIAL;
 
 	key.data = &nid;
+	rc = db->cursor( db, txn, &cursor, bdb->bi_db_opflags );
+	if ( rc ) return rc;
 
 	d = op->o_tmpalloc( data.size, op->o_tmpmemctx );
 	d->nrdnlen[1] = BEI(e)->bei_nrdn.bv_len & 0xff;
@@ -670,13 +593,6 @@ hdb_dn2id_delete(
 	strcpy( d->nrdn, BEI(e)->bei_nrdn.bv_val );
 	data.data = d;
 
-	rc = db->cursor( db, txn, &cursor, bdb->bi_db_opflags );
-	if ( rc ) goto leave;
-
-	/* We hold this lock until the TXN completes */
-	rc = bdb_dn2id_lock( bdb, &e->e_nname, 1, TXN_ID( txn ), &lock );
-	if ( rc ) goto nolock;
-
 	/* Delete our ID from the parent's list */
 	rc = cursor->c_get( cursor, &key, &data, DB_GET_BOTH_RANGE );
 	if ( rc == 0 ) {
@@ -697,10 +613,7 @@ hdb_dn2id_delete(
 		if ( rc == 0 )
 			rc = cursor->c_del( cursor, 0 );
 	}
-
-nolock:
 	cursor->c_close( cursor );
-leave:
 	op->o_tmpfree( d, op->o_tmpmemctx );
 
 	/* Delete IDL cache entries */
@@ -718,7 +631,6 @@ leave:
 			bdb_idl_cache_del_id( bdb, db, &key, e->e_id );
 		}
 	}
-	Debug( LDAP_DEBUG_TRACE, "<= hdb_dn2id_delete 0x%lx: %d\n", e->e_id, rc, 0 );
 	return rc;
 }
 
@@ -726,10 +638,9 @@ leave:
 int
 hdb_dn2id(
 	Operation	*op,
+	DB_TXN *txn,
 	struct berval	*in,
-	EntryInfo	*ei,
-	u_int32_t locker,
-	DB_LOCK *lock )
+	EntryInfo	*ei )
 {
 	struct bdb_info *bdb = (struct bdb_info *) op->o_bd->be_private;
 	DB *db = bdb->bi_dn2id->bdi_db;
@@ -741,8 +652,6 @@ hdb_dn2id(
 	unsigned char dlen[2];
 	ID idp, parentID;
 
-	Debug( LDAP_DEBUG_TRACE, "=> hdb_dn2id(\"%s\")\n", in->bv_val, 0, 0 );
-
 	nrlen = dn_rdnlen( op->o_bd, in );
 	if (!nrlen) nrlen = in->bv_len;
 
@@ -760,11 +669,8 @@ hdb_dn2id(
 	data.dlen = data.ulen;
 	data.flags = DB_DBT_USERMEM | DB_DBT_PARTIAL;
 
-	rc = db->cursor( db, NULL, &cursor, bdb->bi_db_opflags );
+	rc = db->cursor( db, txn, &cursor, bdb->bi_db_opflags );
 	if ( rc ) return rc;
-	if ( locker ) {
-		cursor->locker = locker;
-	}
 
 	d = op->o_tmpalloc( data.size * 3, op->o_tmpmemctx );
 	d->nrdnlen[1] = nrlen & 0xff;
@@ -775,9 +681,6 @@ hdb_dn2id(
 	*ptr = '\0';
 	data.data = d;
 
-	rc = bdb_dn2id_lock( bdb, in, 0, locker, lock );
-	if ( rc ) goto leave;
-
 	rc = cursor->c_get( cursor, &key, &data, DB_GET_BOTH_RANGE );
 	if ( rc == 0 && (dlen[1] != d->nrdnlen[1] || dlen[0] != d->nrdnlen[0] ||
 		strncmp( d->nrdn, in->bv_val, nrlen ))) {
@@ -799,17 +702,8 @@ hdb_dn2id(
 			ei->bei_parent->bei_dkids = dkids;
 		}
 	}
-
-leave:
 	cursor->c_close( cursor );
 	op->o_tmpfree( d, op->o_tmpmemctx );
-	if( rc != 0 ) {
-		Debug( LDAP_DEBUG_TRACE, "<= hdb_dn2id: get failed: %s (%d)\n",
-			db_strerror( rc ), rc, 0 );
-	} else {
-		Debug( LDAP_DEBUG_TRACE, "<= hdb_dn2id: got id=0x%lx\n",
-			ei->bei_id, 0, 0 );
-	}
 
 	return rc;
 }
@@ -817,7 +711,8 @@ leave:
 int
 hdb_dn2id_parent(
 	Operation *op,
-	u_int32_t	locker,
+	DB_TXN *txn,
+	BDB_LOCKER	locker,
 	EntryInfo *ei,
 	ID *idp )
 {
@@ -840,10 +735,10 @@ hdb_dn2id_parent(
 	DBTzero(&data);
 	data.flags = DB_DBT_USERMEM;
 
-	rc = db->cursor( db, NULL, &cursor, bdb->bi_db_opflags );
+	rc = db->cursor( db, txn, &cursor, bdb->bi_db_opflags );
 	if ( rc ) return rc;
-	if ( locker ) {
-		cursor->locker = locker;
+	if ( !txn && locker ) {
+		CURSOR_SETLOCKER(cursor, locker);
 	}
 
 	data.ulen = sizeof(diskNode) + (SLAP_LDAPDN_MAXLEN * 2);
@@ -938,6 +833,7 @@ hdb_dn2id_children(
 struct dn2id_cookie {
 	struct bdb_info *bdb;
 	Operation *op;
+	BDB_LOCKER locker;
 	EntryInfo *ei;
 	ID *ids;
 	ID *tmp;
@@ -1168,7 +1064,9 @@ gotit:
 int
 hdb_dn2idl(
 	Operation	*op,
-	Entry		*e,
+	BDB_LOCKER locker,
+	struct berval *ndn,
+	EntryInfo	*ei,
 	ID *ids,
 	ID *stack )
 {
@@ -1176,20 +1074,21 @@ hdb_dn2idl(
 	struct dn2id_cookie cx;
 
 	Debug( LDAP_DEBUG_TRACE, "=> hdb_dn2idl(\"%s\")\n",
-		e->e_nname.bv_val, 0, 0 );
+		ndn->bv_val, 0, 0 );
 
 #ifndef BDB_MULTIPLE_SUFFIXES
 	if ( op->ors_scope != LDAP_SCOPE_ONELEVEL && 
-		BEI(e)->bei_parent->bei_id == 0 )
+		( ei->bei_id == 0 ||
+		ei->bei_parent->bei_id == 0 ))
 	{
 		BDB_IDL_ALL( bdb, ids );
 		return 0;
 	}
 #endif
 
-	cx.id = e->e_id;
+	cx.id = ei->bei_id;
 	BDB_ID2DISK( cx.id, &cx.nid );
-	cx.ei = e->e_id ? BEI(e) : &bdb->bi_cache.c_dntree;
+	cx.ei = ei;
 	cx.bdb = bdb;
 	cx.db = cx.bdb->bi_dn2id->bdi_db;
 	cx.prefix = (op->ors_scope == LDAP_SCOPE_ONELEVEL) ?
@@ -1198,6 +1097,7 @@ hdb_dn2idl(
 	cx.tmp = stack;
 	cx.buf = stack + BDB_IDL_UM_SIZE;
 	cx.op = op;
+	cx.locker = locker;
 	cx.need_sort = 0;
 	cx.depth = 0;
 
@@ -1225,8 +1125,9 @@ hdb_dn2idl(
 		cx.key.data = ptr;
 		cx.key.size = sizeof(ID)+1;
 		*ptr = cx.prefix;
-		cx.id = e->e_id;
-		bdb_idl_cache_put( cx.bdb, cx.db, &cx.key, cx.ids, cx.rc );
+		cx.id = ei->bei_id;
+		if ( cx.bdb->bi_idl_cache_max_size )
+			bdb_idl_cache_put( cx.bdb, cx.db, &cx.key, cx.ids, cx.rc );
 	}
 
 	if ( cx.rc == DB_NOTFOUND )
