@@ -1,8 +1,8 @@
 /* init.c - initialize various things */
-/* $OpenLDAP: pkg/ldap/servers/slapd/init.c,v 1.81.2.14 2006/10/05 23:31:19 quanah Exp $ */
+/* $OpenLDAP: pkg/ldap/servers/slapd/init.c,v 1.97.2.6 2007/09/27 21:38:25 ralf Exp $ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 1998-2006 The OpenLDAP Foundation.
+ * Copyright 1998-2007 The OpenLDAP Foundation.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -34,9 +34,6 @@
 
 #include "slap.h"
 #include "lber_pvt.h"
-#ifdef LDAP_SLAPI
-#include "slapi/slapi.h"
-#endif
 
 #include "ldap_rq.h"
 
@@ -68,19 +65,10 @@ struct berval NoAttrs = BER_BVC( LDAP_NO_ATTRS );
 ldap_pvt_thread_pool_t	connection_pool;
 int			connection_pool_max = SLAP_MAX_WORKER_THREADS;
 int		slap_tool_thread_max = 1;
-#ifndef HAVE_GMTIME_R
 ldap_pvt_thread_mutex_t	gmtime_mutex;
-#endif
 
 slap_counters_t			slap_counters;
 
-/*
- * these mutexes must be used when calling the entry2str()
- * routine since it returns a pointer to static data.
- */
-ldap_pvt_thread_mutex_t	entry2str_mutex;
-ldap_pvt_thread_mutex_t	replog_mutex;
-
 static const char* slap_name = NULL;
 int slapMode = SLAP_UNDEFINED_MODE;
 
@@ -94,7 +82,7 @@ slap_init( int mode, const char *name )
 
 	if ( slapMode != SLAP_UNDEFINED_MODE ) {
 		/* Make sure we write something to stderr */
-		ldap_debug |= 1;
+		slap_debug |= LDAP_DEBUG_NONE;
 		Debug( LDAP_DEBUG_ANY,
 		 "%s init: init called twice (old=%d, new=%d)\n",
 		 name, slapMode, mode );
@@ -104,9 +92,11 @@ slap_init( int mode, const char *name )
 
 	slapMode = mode;
 
+	slap_op_init();
+
 #ifdef SLAPD_MODULES
 	if ( module_init() != 0 ) {
-		ldap_debug |= 1;
+		slap_debug |= LDAP_DEBUG_NONE;
 		Debug( LDAP_DEBUG_ANY,
 		    "%s: module_init failed\n",
 			name, 0, 0 );
@@ -115,16 +105,32 @@ slap_init( int mode, const char *name )
 #endif
 
 	if ( slap_schema_init( ) != 0 ) {
-		ldap_debug |= 1;
+		slap_debug |= LDAP_DEBUG_NONE;
 		Debug( LDAP_DEBUG_ANY,
 		    "%s: slap_schema_init failed\n",
 		    name, 0, 0 );
 		return 1;
 	}
 
+	if ( filter_init() != 0 ) {
+		slap_debug |= LDAP_DEBUG_NONE;
+		Debug( LDAP_DEBUG_ANY,
+		    "%s: filter_init failed\n",
+		    name, 0, 0 );
+		return 1;
+	}
+
+	if ( entry_init() != 0 ) {
+		slap_debug |= LDAP_DEBUG_NONE;
+		Debug( LDAP_DEBUG_ANY,
+		    "%s: entry_init failed\n",
+		    name, 0, 0 );
+		return 1;
+	}
 
 	switch ( slapMode & SLAP_MODE ) {
 	case SLAP_SERVER_MODE:
+		root_dse_init();
 
 		/* FALLTHRU */
 	case SLAP_TOOL_MODE:
@@ -137,8 +143,6 @@ slap_init( int mode, const char *name )
 
 		ldap_pvt_thread_pool_init( &connection_pool,
 				connection_pool_max, 0);
-		ldap_pvt_thread_mutex_init( &entry2str_mutex );
-		ldap_pvt_thread_mutex_init( &replog_mutex );
 
 		ldap_pvt_thread_mutex_init( &slap_counters.sc_sent_mutex );
 		ldap_pvt_thread_mutex_init( &slap_counters.sc_ops_mutex );
@@ -161,9 +165,7 @@ slap_init( int mode, const char *name )
 		}
 #endif /* SLAPD_MONITOR */
 
-#ifndef HAVE_GMTIME_R
 		ldap_pvt_thread_mutex_init( &gmtime_mutex );
-#endif
 		slap_passwd_init();
 
 		rc = slap_sasl_init();
@@ -177,7 +179,7 @@ slap_init( int mode, const char *name )
 		break;
 
 	default:
-		ldap_debug |= 1;
+		slap_debug |= LDAP_DEBUG_NONE;
 		Debug( LDAP_DEBUG_ANY,
 			"%s init: undefined mode (%d).\n", name, mode, 0 );
 
@@ -186,7 +188,7 @@ slap_init( int mode, const char *name )
 	}
 
 	if ( slap_controls_init( ) != 0 ) {
-		ldap_debug |= 1;
+		slap_debug |= LDAP_DEBUG_NONE;
 		Debug( LDAP_DEBUG_ANY,
 		    "%s: slap_controls_init failed\n",
 		    name, 0, 0 );
@@ -204,7 +206,7 @@ slap_init( int mode, const char *name )
 #endif
 
 	if ( frontend_init() ) {
-		ldap_debug |= 1;
+		slap_debug |= LDAP_DEBUG_NONE;
 		Debug( LDAP_DEBUG_ANY,
 		    "%s: frontend_init failed\n",
 		    name, 0, 0 );
@@ -212,7 +214,7 @@ slap_init( int mode, const char *name )
 	}
 
 	if ( overlay_init() ) {
-		ldap_debug |= 1;
+		slap_debug |= LDAP_DEBUG_NONE;
 		Debug( LDAP_DEBUG_ANY,
 		    "%s: overlay_init failed\n",
 		    name, 0, 0 );
@@ -220,7 +222,7 @@ slap_init( int mode, const char *name )
 	}
 
 	if ( glue_sub_init() ) {
-		ldap_debug |= 1;
+		slap_debug |= LDAP_DEBUG_NONE;
 		Debug( LDAP_DEBUG_ANY,
 		    "%s: glue/subordinate init failed\n",
 		    name, 0, 0 );
@@ -229,7 +231,7 @@ slap_init( int mode, const char *name )
 	}
 
 	if ( acl_init() ) {
-		ldap_debug |= 1;
+		slap_debug |= LDAP_DEBUG_NONE;
 		Debug( LDAP_DEBUG_ANY,
 		    "%s: acl_init failed\n",
 		    name, 0, 0 );
@@ -241,50 +243,22 @@ slap_init( int mode, const char *name )
 
 int slap_startup( Backend *be )
 {
-	int rc;
-
 	Debug( LDAP_DEBUG_TRACE,
 		"%s startup: initiated.\n",
 		slap_name, 0, 0 );
 
 
-	rc = backend_startup( be );
-
-#ifdef LDAP_SLAPI
-	if( rc == 0 ) {
-		Slapi_PBlock *pb = slapi_pblock_new();
-
-		if ( slapi_int_call_plugins( frontendDB, SLAPI_PLUGIN_START_FN, pb ) < 0 ) {
-			rc = -1;
-		}
-		slapi_pblock_destroy( pb );
-	}
-#endif /* LDAP_SLAPI */
-
-	return rc;
+	return backend_startup( be );
 }
 
 int slap_shutdown( Backend *be )
 {
-	int rc;
-#ifdef LDAP_SLAPI
-	Slapi_PBlock *pb;
-#endif
-
 	Debug( LDAP_DEBUG_TRACE,
 		"%s shutdown: initiated\n",
 		slap_name, 0, 0 );
 
 	/* let backends do whatever cleanup they need to do */
-	rc = backend_shutdown( be ); 
-
-#ifdef LDAP_SLAPI
-	pb = slapi_pblock_new();
-	(void) slapi_int_call_plugins( frontendDB, SLAPI_PLUGIN_CLOSE_FN, pb );
-	slapi_pblock_destroy( pb );
-#endif /* LDAP_SLAPI */
-
-	return rc;
+	return backend_shutdown( be ); 
 }
 
 int slap_destroy(void)
@@ -300,14 +274,21 @@ int slap_destroy(void)
 		ber_bvarray_free( default_referral );
 	}
 
+	/* clear out any thread-keys for the main thread */
+	ldap_pvt_thread_pool_context_reset( ldap_pvt_thread_pool_context());
+
 	rc = backend_destroy();
 
 	slap_sasl_destroy();
 
+	/* rootdse destroy goes before entry_destroy()
+	 * because it may use entry_free() */
+	root_dse_destroy();
 	entry_destroy();
 
 	switch ( slapMode & SLAP_MODE ) {
 	case SLAP_SERVER_MODE:
+
 	case SLAP_TOOL_MODE:
 
 		ldap_pvt_thread_mutex_destroy( &slap_counters.sc_sent_mutex );
@@ -336,8 +317,10 @@ int slap_destroy(void)
 
 	}
 
+	slap_op_destroy();
+
 	ldap_pvt_thread_destroy();
 
-	/* should destory the above mutex */
+	/* should destroy the above mutex */
 	return rc;
 }
