@@ -1,5 +1,5 @@
 /* result.c - routines to send ldap results, errors, and referrals */
-/* $OpenLDAP: pkg/ldap/servers/slapd/result.c,v 1.244.2.24 2007/06/13 14:31:47 ralf Exp $ */
+/* $OpenLDAP: pkg/ldap/servers/slapd/result.c,v 1.289.2.7 2007/10/18 00:39:22 quanah Exp $ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
  * Copyright 1998-2007 The OpenLDAP Foundation.
@@ -54,14 +54,6 @@ int slap_freeself_cb( Operation *op, SlapReply *rs )
 	return SLAP_CB_CONTINUE;
 }
 
-int slap_replog_cb( Operation *op, SlapReply *rs )
-{
-	if ( rs->sr_err == LDAP_SUCCESS ) {
-		replog( op );
-	}
-	return SLAP_CB_CONTINUE;
-}
-
 static char *v2ref( BerVarray ref, const char *text )
 {
 	size_t len = 0, i = 0;
@@ -166,7 +158,7 @@ static long send_ldap_ber(
 			return 0;
 		}
 
-		if ( ber_flush( conn->c_sb, ber, 0 ) == 0 ) {
+		if ( ber_flush2( conn->c_sb, ber, LBER_FLUSH_FREE_NEVER ) == 0 ) {
 			break;
 		}
 
@@ -178,7 +170,7 @@ static long send_ldap_ber(
 		 * it's a hard error and return.
 		 */
 
-		Debug( LDAP_DEBUG_CONNS, "ber_flush failed errno=%d reason=\"%s\"\n",
+		Debug( LDAP_DEBUG_CONNS, "ber_flush2 failed errno=%d reason=\"%s\"\n",
 		    err, sock_errstr(err), 0 );
 
 		if ( err != EWOULDBLOCK && err != EAGAIN ) {
@@ -247,7 +239,7 @@ send_ldap_controls( Operation *o, BerElement *ber, LDAPControl **c )
 		if( rc == -1 ) return rc;
 	}
 
-#ifdef LDAP_DEVEL
+#ifdef SLAP_CONTROL_X_SORTEDRESULTS
 	/* this is a hack to avoid having to modify op->s_ctrls */
 	if( o->o_sortedresults ) {
 		BerElementBuffer berbuf;
@@ -277,6 +269,101 @@ send_ldap_controls( Operation *o, BerElement *ber, LDAPControl **c )
 	return rc;
 }
 
+/*
+ * slap_response_play()
+ *
+ * plays the callback list; rationale: a callback can
+ *   - remove itself from the list, by setting op->o_callback = NULL;
+ *     malloc()'ed callbacks should free themselves from inside the
+ *     sc_response() function.
+ *   - replace itself with another (list of) callback(s), by setting
+ *     op->o_callback = a new (list of) callback(s); in this case, it
+ *     is the callback's responsibility to to append existing subsequent
+ *     callbacks to the end of the list that is passed to the sc_response()
+ *     function.
+ *   - modify the list of subsequent callbacks by modifying the value
+ *     of the sc_next field from inside the sc_response() function; this
+ *     case does not require any handling from inside slap_response_play()
+ *
+ * To stop execution of the playlist, the sc_response() function must return
+ * a value different from SLAP_SC_CONTINUE.
+ *
+ * The same applies to slap_cleanup_play(); only, there is no means to stop
+ * execution of the playlist, since all cleanup functions must be called.
+ */
+static int
+slap_response_play(
+	Operation *op,
+	SlapReply *rs )
+{
+	int rc;
+
+	slap_callback	*sc = op->o_callback, **scp;
+
+	rc = SLAP_CB_CONTINUE;
+	for ( scp = &sc; *scp; ) {
+		slap_callback *sc_next = (*scp)->sc_next, **sc_nextp = &(*scp)->sc_next;
+
+		op->o_callback = *scp;
+		if ( op->o_callback->sc_response ) {
+			rc = op->o_callback->sc_response( op, rs );
+			if ( op->o_callback == NULL ) {
+				/* the callback has been removed;
+				 * repair the list */
+				*scp = sc_next;
+				sc_nextp = scp;
+
+			} else if ( op->o_callback != *scp ) {
+				/* a new callback has been inserted
+				 * in place of the existing one; repair the list */
+				*scp = op->o_callback;
+				sc_nextp = scp;
+			}
+			if ( rc != SLAP_CB_CONTINUE ) break;
+		}
+		scp = sc_nextp;
+	}
+
+	op->o_callback = sc;
+	return rc;
+}
+
+static int
+slap_cleanup_play(
+	Operation *op,
+	SlapReply *rs )
+{
+	slap_callback	*sc = op->o_callback, **scp;
+
+	for ( scp = &sc; *scp; ) {
+		slap_callback *sc_next = (*scp)->sc_next, **sc_nextp = &(*scp)->sc_next;
+
+		op->o_callback = *scp;
+		if ( op->o_callback->sc_cleanup ) {
+			(void)op->o_callback->sc_cleanup( op, rs );
+			if ( op->o_callback == NULL ) {
+				/* the callback has been removed;
+				 * repair the list */
+				*scp = sc_next;
+				sc_nextp = scp;
+
+			} else if ( op->o_callback != *scp ) {
+				/* a new callback has been inserted
+				 * after the existing one; repair the list */
+				/* a new callback has been inserted
+				 * in place of the existing one; repair the list */
+				*scp = op->o_callback;
+				sc_nextp = scp;
+			}
+			/* don't care about the result; do all cleanup */
+		}
+		scp = sc_nextp;
+	}
+
+	op->o_callback = sc;
+	return LDAP_SUCCESS;
+}
+
 static int
 send_ldap_response(
 	Operation *op,
@@ -293,25 +380,10 @@ send_ldap_response(
 	}
 
 	if ( op->o_callback ) {
-		slap_callback	*sc = op->o_callback, **sc_prev = &sc, *sc_next;
-
-		rc = SLAP_CB_CONTINUE;
-		for ( sc_next = op->o_callback; sc_next; op->o_callback = sc_next) {
-			sc_next = op->o_callback->sc_next;
-			if ( op->o_callback->sc_response ) {
-				slap_callback *sc2 = op->o_callback;
-				rc = op->o_callback->sc_response( op, rs );
-				if ( op->o_callback != sc2 ) {
-					*sc_prev = op->o_callback;
-				}
-				if ( rc != SLAP_CB_CONTINUE || !op->o_callback ) break;
-				if ( op->o_callback != sc2 ) continue;
-			}
-			sc_prev = &op->o_callback->sc_next;
+		rc = slap_response_play( op, rs );
+		if ( rc != SLAP_CB_CONTINUE ) {
+			goto clean2;
 		}
-
-		op->o_callback = sc;
-		if ( rc != SLAP_CB_CONTINUE ) goto clean2;
 	}
 
 #ifdef LDAP_CONNECTIONLESS
@@ -449,25 +521,9 @@ cleanup:;
 
 clean2:;
 	if ( op->o_callback ) {
-		slap_callback	*sc = op->o_callback, **sc_prev = &sc, *sc_next;
-
-		for ( sc_next = op->o_callback; sc_next; op->o_callback = sc_next) {
-			sc_next = op->o_callback->sc_next;
-			if ( op->o_callback->sc_cleanup ) {
-				slap_callback *sc2 = op->o_callback;
-				(void)op->o_callback->sc_cleanup( op, rs );
-				if ( op->o_callback != sc2 ) {
-					*sc_prev = op->o_callback;
-				}
-				if ( !op->o_callback ) break;
-				if ( op->o_callback != sc2 ) continue;
-			}
-			sc_prev = &op->o_callback->sc_next;
-		}
-		op->o_callback = sc;
+		(void)slap_cleanup_play( op, rs );
 	}
 
-
 	if ( rs->sr_matched && rs->sr_flags & REP_MATCHED_MUSTBEFREED ) {
 		free( (char *)rs->sr_matched );
 		rs->sr_matched = NULL;
@@ -506,7 +562,7 @@ send_ldap_disconnect( Operation	*op, SlapReply *rs )
 	} else {
 		rs->sr_rspoid = LDAP_NOTICE_DISCONNECT;
 		rs->sr_tag = LDAP_RES_EXTENDED;
-		rs->sr_msgid = 0;
+		rs->sr_msgid = LDAP_RES_UNSOLICITED;
 	}
 
 	if ( send_ldap_response( op, rs ) == SLAP_CB_CONTINUE ) {
@@ -687,6 +743,13 @@ slap_send_search_entry( Operation *op, SlapReply *rs )
 		return LDAP_SIZELIMIT_EXCEEDED;
 	}
 
+	/* Every 64 entries, check for thread pool pause */
+	if (( rs->sr_nentries & 0x3f == 0x3f ) &&
+		ldap_pvt_thread_pool_query( &connection_pool,
+			LDAP_PVT_THREAD_POOL_PARAM_PAUSING, &i ) == 0 && i ) {
+		return LDAP_BUSY;
+	}
+
 	rs->sr_type = REP_SEARCH;
 
 	/* eventually will loop through generated operational attribute types
@@ -695,7 +758,7 @@ slap_send_search_entry( Operation *op, SlapReply *rs )
 	/* NOTE: moved before overlays callback circling because
 	 * they may modify entry and other stuff in rs */
 	/* check for special all operational attributes ("+") type */
-	/* FIXME: maybe we could se this flag at the operation level;
+	/* FIXME: maybe we could set this flag at the operation level;
 	 * however, in principle the caller of send_search_entry() may
 	 * change the attribute list at each call */
 	rs->sr_attr_flags = slap_attr_flags( rs->sr_attrs );
@@ -706,26 +769,10 @@ slap_send_search_entry( Operation *op, SlapReply *rs )
 	}
 
 	if ( op->o_callback ) {
-		slap_callback	*sc = op->o_callback, **sc_prev = &sc, *sc_next;
-
-		rc = SLAP_CB_CONTINUE;
-		for ( sc_next = op->o_callback; sc_next; op->o_callback = sc_next )
-		{
-			sc_next = op->o_callback->sc_next;
-			if ( op->o_callback->sc_response ) {
-				slap_callback *sc2 = op->o_callback;
-				rc = op->o_callback->sc_response( op, rs );
-				if ( op->o_callback != sc2 ) {
-					*sc_prev = op->o_callback;
-				}
-				if ( rc != SLAP_CB_CONTINUE || !op->o_callback ) break;
-				if ( op->o_callback != sc2 ) continue;
-			}
-			sc_prev = &op->o_callback->sc_next;
+		rc = slap_response_play( op, rs );
+		if ( rc != SLAP_CB_CONTINUE ) {
+			goto error_return;
 		}
-
-		op->o_callback = sc;
-		if ( rc != SLAP_CB_CONTINUE ) goto error_return;
 	}
 
 	Debug( LDAP_DEBUG_TRACE, "=> send_search_entry: conn %lu dn=\"%s\"%s\n",
@@ -1164,22 +1211,7 @@ slap_send_search_entry( Operation *op, SlapReply *rs )
 
 error_return:;
 	if ( op->o_callback ) {
-		slap_callback	*sc = op->o_callback, **sc_prev = &sc, *sc_next;
-
-		for ( sc_next = op->o_callback; sc_next; op->o_callback = sc_next) {
-			sc_next = op->o_callback->sc_next;
-			if ( op->o_callback->sc_cleanup ) {
-				slap_callback *sc2 = op->o_callback;
-				(void)op->o_callback->sc_cleanup( op, rs );
-				if ( op->o_callback != sc2 ) {
-					*sc_prev = op->o_callback;
-				}
-				if ( !op->o_callback ) break;
-				if ( op->o_callback != sc2 ) continue;
-			}
-			sc_prev = &op->o_callback->sc_next;
-		}
-		op->o_callback = sc;
+		(void)slap_cleanup_play( op, rs );
 	}
 
 	if ( e_flags ) {
@@ -1223,25 +1255,10 @@ slap_send_search_reference( Operation *op, SlapReply *rs )
 
 	rs->sr_type = REP_SEARCHREF;
 	if ( op->o_callback ) {
-		slap_callback	*sc = op->o_callback, **sc_prev = &sc, *sc_next;
-
-		rc = SLAP_CB_CONTINUE;
-		for ( sc_next = op->o_callback; sc_next; op->o_callback = sc_next) {
-			sc_next = op->o_callback->sc_next;
-			if ( op->o_callback->sc_response ) {
-				slap_callback *sc2 = op->o_callback;
-				rc = op->o_callback->sc_response( op, rs );
-				if ( op->o_callback != sc2 ) {
-					*sc_prev = op->o_callback;
-				}
-				if ( rc != SLAP_CB_CONTINUE || !op->o_callback ) break;
-				if ( op->o_callback != sc2 ) continue;
-			}
-			sc_prev = &op->o_callback->sc_next;
+		rc = slap_response_play( op, rs );
+		if ( rc != SLAP_CB_CONTINUE ) {
+			goto rel;
 		}
-
-		op->o_callback = sc;
-		if ( rc != SLAP_CB_CONTINUE ) goto rel;
 	}
 
 	Debug( LDAP_DEBUG_TRACE,
@@ -1371,22 +1388,7 @@ slap_send_search_reference( Operation *op, SlapReply *rs )
 
 rel:
 	if ( op->o_callback ) {
-		slap_callback	*sc = op->o_callback, **sc_prev = &sc, *sc_next;
-
-		for ( sc_next = op->o_callback; sc_next; op->o_callback = sc_next) {
-			sc_next = op->o_callback->sc_next;
-			if ( op->o_callback->sc_cleanup ) {
-				slap_callback *sc2 = op->o_callback;
-				(void)op->o_callback->sc_cleanup( op, rs );
-				if ( op->o_callback != sc2 ) {
-					*sc_prev = op->o_callback;
-				}
-				if ( rc != SLAP_CB_CONTINUE || !op->o_callback ) break;
-				if ( op->o_callback != sc2 ) continue;
-			}
-			sc_prev = &op->o_callback->sc_next;
-		}
-		op->o_callback = sc;
+		(void)slap_cleanup_play( op, rs );
 	}
 
 	return rc;
@@ -1502,14 +1504,15 @@ int slap_read_controls(
 		op->o_preread_attrs : op->o_postread_attrs; 
 
 	bv.bv_len = entry_flatsize( rs->sr_entry, 0 );
-	bv.bv_val = op->o_tmpalloc(bv.bv_len, op->o_tmpmemctx );
+	bv.bv_val = op->o_tmpalloc( bv.bv_len, op->o_tmpmemctx );
 
 	ber_init2( ber, &bv, LBER_USE_DER );
 	ber_set_option( ber, LBER_OPT_BER_MEMCTX, &op->o_tmpmemctx );
 
 	/* create new operation */
 	myop = *op;
-	myop.o_bd = NULL;
+	/* FIXME: o_bd needed for ACL */
+	myop.o_bd = op->o_bd;
 	myop.o_res_ber = ber;
 	myop.o_callback = NULL;
 	myop.ors_slimit = 1;
